/*
Copyright (c) 2015, Raspberry Pi Foundation
Copyright (c) 2015, Dave Stevenson
Copyright (c) 2020-2021, Soho enterprise
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <bcm_host.h>

static const char __PROGRAM_VERSION__[] = "@(#)cap.c Ver. 1.0022 (2021/10/03 22:08:45)";

#define IMAGE_ID_10    0x2B
#define IMAGE_ID_12    0x2C

#include "interface/mmal/mmal.h"
#include "interface/mmal/util/mmal_default_components.h"
#include "interface/mmal/util/mmal_util.h"
#include "interface/mmal/util/mmal_util_params.h"

static unsigned char *RawBuf;
static int      RawReadFlag = 0;
static int      Width  = 0;
static int      Height = 0;
static int      BufSiz = 0;
static int      Size = 0;
static int      running = 0;

typedef struct {
    int write_empty;
} RASPIRAW_PARAMS_T;

RASPIRAW_PARAMS_T cfg = {
    .write_empty = 0,
};

#define  ERR_                   (-1)
#define  ERR_component_destroy  (-2)
#define  ERR_component_disable  (-3)
#define  ERR_pool_destroy       (-4)
#define  ERR_port_disable       (-5)

int     InitSensor(char *fname, int *width, int *height, int *bufsiz);
void    StopSensor();
int     check_lanes(char *fname);
int     check_rawbit(char *fname);

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
static void callback(MMAL_PORT_T *port, MMAL_BUFFER_HEADER_T *buffer)
{
    if (!running) {
        mmal_buffer_header_release(buffer);
    } else {
        RASPIRAW_PARAMS_T *cfg = (RASPIRAW_PARAMS_T *)port->userdata;

        if (!(buffer->flags&MMAL_BUFFER_HEADER_FLAG_CODECSIDEINFO)) {
            if (buffer->length != BufSiz) {
                fprintf(stderr, "BUFSIZ@callback = %d\n", (int)buffer->length);
                fflush(stderr);
            }
            if (!cfg->write_empty) {
//-------------------------------------------------------------
                if ((RawBuf != NULL) && (RawReadFlag == 0)) {
                    memcpy((void *)RawBuf, (void *)buffer->data, (int)buffer->length);
                    RawReadFlag = 1;
                }
//-------------------------------------------------------------
            }
        }

        buffer->length = 0;
        mmal_port_send_buffer(port, buffer);
    }
}

MMAL_COMPONENT_T *rawcam=NULL, *isp=NULL, *render=NULL;
MMAL_PORT_T *output = NULL;
MMAL_POOL_T *pool = NULL;
MMAL_PARAMETER_CAMERA_RX_CONFIG_T rx_cfg = {{MMAL_PARAMETER_CAMERA_RX_CONFIG, sizeof(rx_cfg)}};
MMAL_PARAMETER_CAMERA_RX_TIMING_T rx_timing = {{MMAL_PARAMETER_CAMERA_RX_TIMING, sizeof(rx_timing)}};
/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
OpenSensor(char *fname)
{
    u_int32_t encoding;

    bcm_host_init();
    bcm_host_init();

    if (mmal_component_create("vc.ril.rawcam", &rawcam) != MMAL_SUCCESS) {
        return ERR_;
    }

    if (mmal_component_create("vc.ril.isp", &isp) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }

    if (mmal_component_create(MMAL_COMPONENT_DEFAULT_VIDEO_RENDERER, &render) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }

    output = rawcam->output[0];
    if (mmal_port_parameter_get(output, &rx_cfg.hdr) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }

    rx_cfg.unpack = MMAL_CAMERA_RX_CONFIG_UNPACK_NONE;
    rx_cfg.pack   = MMAL_CAMERA_RX_CONFIG_PACK_NONE;

    rx_cfg.data_lanes = check_lanes(fname);  // DATA_LANES
    if (check_rawbit(fname) == 12) {
        rx_cfg.image_id   = IMAGE_ID_12;
        encoding = MMAL_ENCODING_BAYER_SBGGR12P;
    } else {
        rx_cfg.image_id   = IMAGE_ID_10;
        encoding = MMAL_ENCODING_BAYER_SBGGR10P;
    }

    if (mmal_port_parameter_set(output, &rx_cfg.hdr) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }
    if (mmal_port_parameter_get(output, &rx_timing.hdr) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }
    if (mmal_port_parameter_set(output, &rx_timing.hdr) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }
    if (mmal_component_enable(rawcam) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }
    if (mmal_component_enable(isp) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }
    if (mmal_component_enable(render) != MMAL_SUCCESS) {
        return  ERR_component_destroy;
    }

    if (InitSensor(fname, &Width, &Height, &BufSiz) < 0) {
        return  ERR_;
    }

    if (BufSiz == 0) {
        Size = 16;
    } else {
        Size = BufSiz;
    }
    RawBuf = (unsigned char *)malloc(Size);
//printf("Width=%d, Height=%d, Size=%d\n", Width, Height, Size);

    output->format->es->video.crop.width  = Width;
    output->format->es->video.crop.height = Height;
    output->format->es->video.width  = VCOS_ALIGN_UP(Width , 16);
    output->format->es->video.height = VCOS_ALIGN_UP(Height, 16);
    output->format->encoding = encoding;

    if (mmal_port_format_commit(output) != MMAL_SUCCESS) {
        return  ERR_component_disable;
    }

    output->buffer_size = output->buffer_size_recommended;
    output->buffer_num  = output->buffer_num_recommended;

    if (mmal_port_parameter_set_boolean(output, MMAL_PARAMETER_ZERO_COPY, MMAL_TRUE) != MMAL_SUCCESS) {
        return  ERR_component_disable;
    }

    pool = mmal_port_pool_create(output, output->buffer_num, output->buffer_size);
    if (!pool) {
        return  ERR_component_disable;
    }

    output->userdata = (struct MMAL_PORT_USERDATA_T *)&cfg;
    if (mmal_port_enable(output, callback) != MMAL_SUCCESS) {
        return  ERR_pool_destroy;
    }

    running = 1;
    for (int i = 0; i < (int)output->buffer_num; i++) {
        MMAL_BUFFER_HEADER_T *buffer;

        if ((buffer = mmal_queue_get(pool->queue)) == NULL) {
            return  ERR_port_disable;
        }
        if (mmal_port_send_buffer(output, buffer) != MMAL_SUCCESS) {
            return  ERR_port_disable;
        }
    }

    return  0;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
CloseSensor(int status)
{
    running = 0;

    switch (status) {
    case ERR_:                  goto component_destroy;
    case ERR_component_destroy: goto component_destroy;
    case ERR_component_disable: goto component_disable;
    case ERR_pool_destroy:      goto pool_destroy;
    case ERR_port_disable:      goto port_disable;
    default:                    break;
    }

    StopSensor();

port_disable:
    if (mmal_port_disable(output) != MMAL_SUCCESS) {
        return -1;
    }

pool_destroy:
    if (pool)
        mmal_port_pool_destroy(output, pool);

component_disable:
    (void)mmal_component_disable(render);
    (void)mmal_component_disable(isp);
    (void)mmal_component_disable(rawcam);

component_destroy:
    if (rawcam)
        mmal_component_destroy(rawcam);
    if (isp)
        mmal_component_destroy(isp);
    if (render)
        mmal_component_destroy(render);
    if (RawBuf)
        free(RawBuf);

    return 0;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
ReadRAW(unsigned char *buf)
{
    if (RawReadFlag == 0) {
        return  0;
    }

    memcpy((void *)buf, (void *)RawBuf, BufSiz);

    RawReadFlag = 0;

    return  1;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
GetWidth()
{
    return  Width;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
GetHeight()
{
    return  Height;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
GetBufSiz()
{
    return  BufSiz;
}
