/*====================================================================*
 * ser.c
 *
 * Copyright(c) 2022 TsukiTsuki. (http://tsukitsuki.tips/)
 *====================================================================*/
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/statvfs.h>
#include <string.h>
#include <math.h>
#include <regex.h>
#include "fits.h"
#include "ser.h"

unsigned char *SerImage;
char *SerHeader;
char sensorname2[18];
int FrameCount;

/*--------------------------------------------------------------------*
 *  SERファイル保存用の各データを初期化する
 *--------------------------------------------------------------------*/
int InitSer(int bpp, int width, int height, char *fname, int colorID)
{
    GetSensorName(fname, sensorname2);

    SerHeader = (char *)malloc(SER_BLOCK_SIZE);
    if(SerHeader == NULL){
        perror("malloc error");
        exit(EXIT_FAILURE);
    }

    char *SHp = SerHeader;
    for(int i=0; i<178; i++) {
        *SHp++ = 0;
    }

    int offset = 0;
    sprintf(SerHeader+offset,"LUCAM-RECODER");
    offset += SER_FILEID_SIZE;
    offset += SER_LUID_SIZE;
//    *(int *)(SerHeader+offset) = 8; //RGGB
    *(int *)(SerHeader+offset) = colorID;
    offset += SER_COLORID_SIZE;
    offset += SER_LITTLEENDIAN_SIZE;
    *(int *)(SerHeader+offset) = width;
    offset += SER_WIDTH_SIZE;
    *(int *)(SerHeader+offset) = height;
    offset += SER_HEIGHT_SIZE;
    *(int *)(SerHeader+offset) = bpp;
    offset += SER_PIXELDPTH_SIZE;
    offset += SER_FRAMES_SIZE;
    sprintf(SerHeader+offset,"%-40s", "Observer");
    offset += SER_OBSERVER_SIZE;
    sprintf(SerHeader+offset,"%-40s", sensorname2);
    offset += SER_INSTRUMENT_SIZE;
    sprintf(SerHeader+offset,"%-40s", "Telescope");
    offset += SER_TELESCOPE_SIZE;
    offset += SER_DATETIME_SIZE;
    offset += SER_DATETIMEUTC_SIZE;

    return 0;
}

FILE *fpSer;
FILE *fpTS;
struct timeval tv;

/*--------------------------------------------------------------------*
 *  SERファイルにヘッダを書き出す
 *--------------------------------------------------------------------*/
int WriteSERHeader(char *wfname)
{
    FrameCount = 0;
    fpSer = fopen64(wfname, "wb");
    fwrite(SerHeader, sizeof(char), SER_BLOCK_SIZE, fpSer);
    printf("Rec Start.\n");

    // タイムスタンプを一旦テンポラリファイルに記録
    fpTS = fopen("/tmp/timestamp.bin", "wb");

    return 0;
}

/*--------------------------------------------------------------------*
 *  SERファイルに画像を書き出す
 *--------------------------------------------------------------------*/
int WriteSERImage(unsigned char *BinImage, int BinSize, int bpp)
{
    gettimeofday(&tv, NULL);
    
//    EndianChange(SerImage, BinImage, BinSize, bpp);
    fwrite(BinImage, sizeof(char), BinSize, fpSer);
    FrameCount++;
    fwrite(&tv, sizeof(tv), 1, fpTS);        // 8bytes

    return 0;
}

/*--------------------------------------------------------------------*
 *  日付から100ナノ秒単位に変換
 *--------------------------------------------------------------------*/
int64_t date2timestamp(int year, int month, int day, int hour, int minute, int second, int microsec){
    int monthdays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}; //平年の月の日数

    // グレゴリオ暦で年を日数に変換
    year += 1900;
    int year2day = (year -1) *365 + year/400 - year/100 + year/4;
    
    // 年内の日数を計算
    int mon2day;
    for(int i=0; i < month; i++){
        mon2day += monthdays[i];
    }
    int days = year2day + mon2day + day -1;

    // 秒に換算
    int64_t seconds = ((int64_t)hour *60 + (int64_t)minute) *60 + (int64_t)second;
    seconds += (int64_t)days * 24 *60 *60;

    // 100ナノ秒に換算
    return (seconds * 1000 * 1000 + (int64_t)microsec) * 10;
}

/*--------------------------------------------------------------------*
 *  SERファイルに画像を終了する
 *--------------------------------------------------------------------*/
int CloseSER()
{
    struct tm *ts;
    int64_t TimeDate;

    // フレーム数を記録
    fseek(fpSer, SER_FRAMES_OFFSET, SEEK_SET);
    fwrite(&FrameCount, 1, sizeof(int), fpSer);

    fclose(fpTS);

    // タイムスタンプの記録ファイルを開く
    fpTS = fopen("/tmp/timestamp.bin", "rb");
    fread(&tv, sizeof(tv), 1, fpTS);
    fclose(fpTS);

    // ローカルタイムを記録
    ts = localtime(&tv.tv_sec);
    TimeDate = date2timestamp(ts->tm_year, ts->tm_mon, ts->tm_mday,
        ts->tm_hour, ts->tm_min, ts->tm_sec, tv.tv_usec);
    fseek(fpSer, SER_DATETIME_OFFSET, SEEK_SET);
    fwrite(&TimeDate, sizeof(TimeDate), 1, fpSer);
    printf("%04d-%02d-%02d %02d:%02d:%02d.%06d JST\n",ts->tm_year +1900, ts->tm_mon +1,
        ts->tm_mday, ts->tm_hour, ts->tm_min, ts->tm_sec, tv.tv_usec);

    // UTCを記録
    ts = gmtime(&tv.tv_sec);
    TimeDate = date2timestamp(ts->tm_year, ts->tm_mon, ts->tm_mday,
        ts->tm_hour, ts->tm_min, ts->tm_sec, tv.tv_usec);
    fseek(fpSer, SER_DATETIMEUTC_OFFSET, SEEK_SET);
    fwrite(&TimeDate, sizeof(TimeDate), 1, fpSer);
    printf("%04d-%02d-%02d %02d:%02d:%02d.%06d UTC\n",ts->tm_year +1900, ts->tm_mon +1,
        ts->tm_mday, ts->tm_hour, ts->tm_min, ts->tm_sec, tv.tv_usec);

    // タイムスタンプをトレーラに記録
    fseek(fpSer, 0, SEEK_END);
    fpTS = fopen("/tmp/timestamp.bin", "rb");
    for(int i=0; i<FrameCount; i++){
        fread(&tv, sizeof(tv), 1, fpTS);
        ts = gmtime(&tv.tv_sec);
        TimeDate = date2timestamp(ts->tm_year, ts->tm_mon, ts->tm_mday,
            ts->tm_hour, ts->tm_min, ts->tm_sec, tv.tv_usec);
        fwrite(&TimeDate, sizeof(TimeDate), 1, fpSer);
    }
    fclose(fpTS);

    fclose(fpSer);
    printf("%dframes\n", FrameCount);
    printf("Rec Stop.\n");

    return 0;
}
